"""Exceptions triggered to warn about internal errors in Assonant Data class module."""


class AssonantDataClassesError(Exception):
    """Custom Exception to warn when any error occurs inside Assonant Data Classes Modules."""



"""Assonant data classes enums.

This submodule defines Enumations classes used to standardize options from some data classes.
"""
from .beamline_name import BeamlineName
from .measurement_type import MeasurementType
from .scope_type import ScopeType
from .transformation_type import TransformationType

__all__ = [
    "BeamlineName",
    "MeasurementType",
    "ScopeType",
    "TransformationType",
]



"""Assonant Entry data class."""
from typing import Dict, List, Optional, Type, Union

import numpy as np

from .assonant_data_class import AssonantDataClass
from .components import Beamline
from .data_handlers import DataField, DataHandler, TimeSeries
from .enums import ScopeType
from .exceptions import AssonantDataClassesError


class Entry(AssonantDataClass):
    """Data classes that wraps data into a logical/temporal scope related to the experiment.

    Entries are used to group and represent data in a defined temporal/logical scope of the
    experiment, which is directly define by the field "scope_type". e.g: calibration,
    pre-exposition. Every Entry instance must also have an Beamline object to identify
    with at least the beamline name it is related to.
    """

    scope_type: ScopeType
    beamline: Beamline
    fields: Optional[Dict[str, DataHandler]] = {}

    def add_field(
        self,
        name: str,
        value: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ):
        """Add new positional field to component that data was collected as a TimeSeries.

        Args:
            name (str): Field name.
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to field
            collected data.
            unit (Optional[str], optional): Measurement unit related to value parameter.
            Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]], optional): Dict
            containing any aditional metadata related to collected data. Defaults to {}.
        """
        try:
            new_field = DataField(
                value=value,
                unit=unit,
                extra_metadata=extra_metadata,
            )
        except Exception as e:
            raise AssonantDataClassesError(f"Failed to create DataField Data Handler for {self.name} Component.") from e
        if name not in self.fields:
            self.fields[name] = new_field
        else:
            raise AssonantDataClassesError(f"Field name already exists on: {self.name} Component.")

    def add_timeseries_field(
        self,
        name: str,
        value: Union[int, float, str, List, Type[np.ndarray], None],
        timestamps: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
        timestamps_unit: Optional[str] = None,
        timestamp_extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ):
        """Add new positional field to component that data was collected as a TimeSeries.

        Args:
            name (str): Field name.
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to field
            collected data.
            timestamps (Union[int, float, str, List, Type[np.ndarray], None]): Timestamps related to data collected
            from the field.
            unit (Optional[str], optional): Measurement unit related to value parameter.
            Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing any aditional metadata related to collected data. Defaults to {}.
            timestamps_unit (Optional[str], optional): Measurement unit related to
            timestamp field. Defaults to None.
            timestamp_extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing extra metadata about timestamps field. Defaults to {}.
        """
        try:
            new_field = TimeSeries(
                value=DataField(
                    value=value,
                    unit=unit,
                    extra_metadata=extra_metadata,
                ),
                timestamps=DataField(
                    value=timestamps,
                    unit=timestamps_unit,
                    extra_metadata=timestamp_extra_metadata,
                ),
            )
        except Exception as e:
            raise AssonantDataClassesError(
                f"Failed to create TimeSeries Data Handler for {self.name} Component."
            ) from e
        if name not in self.fields:
            self.fields[name] = new_field
        else:
            raise AssonantDataClassesError(f"Field name already exists on: {self.name} Component.")



"""Assonant Beamline component data class."""

from ..enums.beamline_name import BeamlineName
from .component import Component


class Beamline(Component):
    """Data class to handle all metadata related to them beamline."""

    name: BeamlineName



"""Assonant Component abstract class."""

from typing import Dict, List, Optional, Type, Union

import numpy as np

from ..assonant_data_class import AssonantDataClass
from ..data_handlers import Axis, DataField, DataHandler, TimeSeries
from ..enums import TransformationType
from ..exceptions import AssonantDataClassesError


# TODO: Make this class abstract
class Component(AssonantDataClass):
    """Abstract class that creates the base common requirements to define an Assonant Component.

    Components are more generic definitions which may be composed by many subcomponents if more
    detailing in its composition is desired.
    """

    name: str
    subcomponents: Optional[List["Component"]] = []
    positions: Optional[List[Axis]] = []
    fields: Optional[Dict[str, DataHandler]] = {}

    def add_subcomponent(self, component: Union["Component", List["Component"]]):
        """Add new subcomponent or list of new subcomponents to component.

        Args:
            component (Union[Component, List[Component]]): Component object or List of Components which will be add as
            subcomponent from called Component object.
        """
        if isinstance(component, List):
            for _component in component:
                self.subcomponents.append(_component)
        else:
            self.subcomponents.append(component)

    def add_position(
        self,
        name: str,
        transformation_type: TransformationType,
        value: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ):
        """Add new positional field to component.

        Args:
            name (str): Axis name.
            transformation_type (TransformationType): Type of transformation done by axis.
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to axis
            collected data.
            unit (Optional[str], optional): Measurement unit related to value parameter.
            Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing any aditional metadata related to collected data. Defaults to {}.
        """
        try:
            new_axis = Axis(
                name=name,
                transformation_type=transformation_type,
                value=DataField(
                    value=value,
                    unit=unit,
                    extra_metadata=extra_metadata,
                ),
            )
        except Exception as e:
            raise AssonantDataClassesError(f"Failed to create Axis Data Handler for {self.name} Component.") from e
        try:
            self.positions.append(new_axis)
        except Exception as e:
            raise AssonantDataClassesError(f"Failed to add Axis to {self.name} Component positions list.") from e

    def add_timeseries_position(
        self,
        name: str,
        transformation_type: TransformationType,
        value: Union[int, float, str, List, Type[np.ndarray], None],
        timestamps: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
        timestamps_unit: Optional[str] = None,
        timestamp_extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ):
        """Add new positional field to component.

        Args:
            name (str): Axis name
            transformation_type (TransformationType): Transformation type of the Axis
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to
            axis collected data
            timestamps (Union[int, float, str, List, Type[np.ndarray], None]): Timestamps
            related to data collected from the axis.
            unit (Optional[str], optional): Measurement unit related to value
            field. Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing extra metadata about value field. Defaults to {}.
            tracked it as a TimeSeries. Defaults to None.
            timestamps_unit (Optional[str], optional): Measurement unit related to
            timestamp field. Defaults to None.
            timestamp_extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dict
            containing extra metadata about timestamps field. Defaults to {}.
        """
        # Check if positions should be saved as a DataField or TimeSeries
        try:
            new_axis = Axis(
                name=name,
                transformation_type=transformation_type,
                value=TimeSeries(
                    value=DataField(
                        value=value,
                        unit=unit,
                        extra_metadata=extra_metadata,
                    ),
                    timestamps=DataField(
                        value=timestamps,
                        unit=timestamps_unit,
                        extra_metadata=timestamp_extra_metadata,
                    ),
                ),
            )
        except Exception as e:
            raise AssonantDataClassesError(
                f"Failed to create Axis Data TimeSeries with TimeSeries data for {self.name} Component."
            ) from e
        try:
            self.positions.append(new_axis)
        except Exception as e:
            raise AssonantDataClassesError(
                f"Failed to add Axis with TimeSeries data to {self.name} Component positions list."
            ) from e

    def add_field(
        self,
        name: str,
        value: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ):
        """Add new positional field to component that data was collected as a TimeSeries.

        Args:
            name (str): Field name.
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to field
            collected data.
            unit (Optional[str], optional): Measurement unit related to value parameter.
            Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]], optional): Dict
            containing any aditional metadata related to collected data. Defaults to {}.
        """
        try:
            new_field = DataField(
                value=value,
                unit=unit,
                extra_metadata=extra_metadata,
            )
        except Exception as e:
            raise AssonantDataClassesError(f"Failed to create DataField Data Handler for {self.name} Component.") from e
        if name not in self.fields:
            self.fields[name] = new_field
        else:
            raise AssonantDataClassesError(f"Field name already exists on: {self.name} Component.")

    def add_timeseries_field(
        self,
        name: str,
        value: Union[int, float, str, List, Type[np.ndarray], None],
        timestamps: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
        timestamps_unit: Optional[str] = None,
        timestamp_extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ):
        """Add new positional field to component that data was collected as a TimeSeries.

        Args:
            name (str): Field name.
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to field
            collected data.
            timestamps (Union[int, float, str, List, Type[np.ndarray], None]): Timestamps related to data collected
            from the field.
            unit (Optional[str], optional): Measurement unit related to value parameter.
            Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing any aditional metadata related to collected data. Defaults to {}.
            timestamps_unit (Optional[str], optional): Measurement unit related to
            timestamp field. Defaults to None.
            timestamp_extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing extra metadata about timestamps field. Defaults to {}.
        """
        try:
            new_field = TimeSeries(
                value=DataField(
                    value=value,
                    unit=unit,
                    extra_metadata=extra_metadata,
                ),
                timestamps=DataField(
                    value=timestamps,
                    unit=timestamps_unit,
                    extra_metadata=timestamp_extra_metadata,
                ),
            )
        except Exception as e:
            raise AssonantDataClassesError(
                f"Failed to create TimeSeries Data Handler for {self.name} Component."
            ) from e
        if name not in self.fields:
            self.fields[name] = new_field
        else:
            raise AssonantDataClassesError(f"Field name already exists on: {self.name} Component.")



"""Scope type definitions."""
from enum import Enum


class MeasurementType(Enum):
    """Measurement types allowed on sensors."""

    TEMPERATURE = "temperature"
    PH = "pH"
    MAGNETIC_FIELD = "magnetic_field"
    ELECTRIC_FIELD = "electric_field"
    CONDUCTIVITY = "conductivity"
    RESISTANCE = "resistance"
    VOLTAGE = "voltage"
    PRESSURE = "pressure"
    FLOW = "flow"
    STRESS = "stress"
    STRAIN = "strain"
    SHEAR = "shear"
    SURFACE_PRESSURE = "surface_pressure"



"""Assonant Sensor component data class."""

from ..enums import MeasurementType
from .component import Component


class Sensor(Component):
    """Data class to handle all data required to define a sensor with a specific measurement.

    This class was created handle measurements that are not necessarly related to a specific device.
    That saidwhen some measurement is done related to the Sample or to a specific device, the
    information should be store inside the specific class for it and not into a Sensor class.
    """

    measurement_type: MeasurementType
